/*
 *	Auxiliary functions for processing PDF files
 *
 *	(c) 2018--2022 Martin Mares <mj@ucw.cz>
 */

#include <cstdio>
#include <cstdlib>
#include <cstring>

#include <iconv.h>

#include "jam.h"

#include <qpdf/QUtil.hh>
#include <qpdf/Pl_Concatenate.hh>

/*** Transformation matrices ***/

// Construct string representation of a transformation matrix
string pdf_matrix::to_string() {
	string s;
	for (int i=0; i<6; i++) {
		if (i)
			s += " ";
		s += pdf_coord(m[i], 6);
	}
	return s;
}

/*** Bounding boxes ***/

QPDFObjectHandle BBox::to_array()
{
	QPDFObjectHandle a = QPDFObjectHandle::newArray();
	a.appendItem(QPDFObjectHandle::newReal(x_min, 1));
	a.appendItem(QPDFObjectHandle::newReal(y_min, 1));
	a.appendItem(QPDFObjectHandle::newReal(x_max, 1));
	a.appendItem(QPDFObjectHandle::newReal(y_max, 1));
	return a;
}

string BBox::to_rect()
{
	return
		pdf_coord(x_min) + " " +
		pdf_coord(y_min) + " " +
		pdf_coord(width()) + " " +
		pdf_coord(height());
}

bool BBox::parse(QPDFObjectHandle h)
{
	if (!h.isArray() || h.getArrayNItems() != 4)
		return false;
	double x[4];
	for (int i=0; i<4; i++) {
		QPDFObjectHandle item = h.getArrayItem(i);
		if (!item.isNumber())
			return false;
		x[i] = item.getNumericValue();
	}
	x_min = x[0];
	y_min = x[1];
	x_max = x[2];
	y_max = x[3];
	return true;
}

BBox::BBox(QPDFObjectHandle box) {
	if (!parse(box)) {
		warn("Invalid bounding box, falling back to A4");
		x_min = 0, x_max = A4_WIDTH;
		y_min = 0, y_max = A4_HEIGHT;
	}
}

void BBox::transform(pdf_matrix &m)
{
	m.apply(&x_min, &y_min);
	m.apply(&x_max, &y_max);
	if (x_max < x_min)
		swap(x_min, x_max);
	if (y_max < y_min)
		swap(y_min, y_max);
}

BBox BBox::transformed(pdf_matrix &m)
{
	BBox b = *this;
	b.transform(m);
	return b;
}

void BBox::join(BBox &with)
{
	x_min = min(x_min, with.x_min);
	x_max = max(x_max, with.x_max);
	y_min = min(y_min, with.y_min);
	y_max = max(y_max, with.y_max);
}

static double clamp(double x, double min, double max)
{
	if (x < min)
		return min;
	if (x > max)
		return max;
	return x;
}

void BBox::intersect(BBox &with)
{
	x_min = clamp(x_min, with.x_min, with.x_max);
	x_max = clamp(x_max, with.x_min, with.x_max);
	y_min = clamp(y_min, with.y_min, with.y_max);
	y_max = clamp(y_max, with.y_min, with.y_max);
}

void BBox::enlarge(double by)
{
	x_min -= by;
	x_max += by;
	y_min -= by;
	y_max += by;
}

BBox BBox::enlarged(double by)
{
	BBox b = *this;
	b.enlarge(by);
	return b;
}

/*** Unicode strings ***/

// Construct PDF representation of a UTF-8 string
QPDFObjectHandle unicode_string(string s)
{
	// If it is ASCII only, use the string directly
	bool ascii_only = true;
	for (char c: s)
		if (c < 0x20 || c > 0x7e)
			ascii_only = false;
	if (ascii_only)
		return QPDFObjectHandle::newString(s);

	// Use iconv to convert the string to big-endian UTF-16
	iconv_t conv = iconv_open("UTF-16BE", "UTF-8");
	if (conv == (iconv_t) -1)
		die("Cannot initialize iconv: %m");

	char *in_ptr = (char *) s.c_str();	// Work around bad API of iconv()
	size_t in_len = strlen(in_ptr);
	size_t out_len = 2*in_len + 2;		// Worst case (including the BOM)
	char out_buf[out_len];
	char *out_ptr = out_buf;
	size_t res = iconv(conv, &in_ptr, &in_len, &out_ptr, &out_len);
	if (res == (size_t) -1)
		die("iconv failed: %m");
	if (in_len)
		die("iconv stopped before the end of input");

	iconv_close(conv);

	// Package UTF-16 in a PDF string
	string out;
	out += 0xfe;
	out += 0xff;
	for (char *p = out_buf; p < out_ptr; p++)
		out += *p;
	return QPDFObjectHandle::newString(out);
}

/*** Conversion of pages to XObjects ***/

static BBox get_trim_box(QPDFObjectHandle page)
{
	static const char * const boxes[] = { "/TrimBox", "/CropBox", "/MediaBox", NULL };
	for (int i=0; boxes[i]; i++)
		if (page.hasKey(boxes[i]))
			return BBox(page.getKey(boxes[i]));
	warn("Page has no trimbox, falling back to A4");
	return BBox(0, 0, A4_WIDTH, A4_HEIGHT);
}

/* Conversion of pages to XObjects is inspired by CUPS's pdftopdf filter. */
class CombineFromContents_Provider : public QPDFObjectHandle::StreamDataProvider {
	private:
		vector<QPDFObjectHandle> contents;
	public:
		CombineFromContents_Provider(const vector<QPDFObjectHandle> &contents) : contents(contents) { }
		void provideStreamData(int objid UNUSED, int generation UNUSED, Pipeline* pipeline) {
			Pl_Concatenate concat("concat", pipeline);
			for (int i=0; i < (int)contents.size(); i++)
				contents[i].pipeStreamData(&concat, true, false, false);
			concat.manualFinish();
		}
};

QPDFObjectHandle page_to_xobject(QPDF *out, QPDFObjectHandle page)
{
	page.assertPageObject();

	QPDFObjectHandle xo_stream = QPDFObjectHandle::newStream(out);
	QPDFObjectHandle xo_dict = xo_stream.getDict();

	xo_dict.replaceKey("/Type", QPDFObjectHandle::newName("/XObject"));
	xo_dict.replaceKey("/Subtype", QPDFObjectHandle::newName("/Form"));
	xo_dict.replaceKey("/FormType", QPDFObjectHandle::newInteger(1));

	BBox box = get_trim_box(page);
	xo_dict.replaceKey("/BBox", box.to_array());

	xo_dict.replaceKey("/Resources", page.getKey("/Resources"));
	if (page.hasKey("/Group"))
		xo_dict.replaceKey("/Group", page.getKey("/Group"));

	if (page.hasKey("/UserUnit")) {
		double u = page.getKey("/UserUnit").getNumericValue();
		QPDFObjectHandle m = QPDFObjectHandle::newArray();
		m.appendItem(QPDFObjectHandle::newReal(u, 3));
		m.appendItem(QPDFObjectHandle::newReal(0, 0));
		m.appendItem(QPDFObjectHandle::newReal(0, 0));
		m.appendItem(QPDFObjectHandle::newReal(u, 3));
		m.appendItem(QPDFObjectHandle::newReal(0, 0));
		m.appendItem(QPDFObjectHandle::newReal(0, 0));
		xo_dict.replaceKey("/Matrix", m);
	}

	vector<QPDFObjectHandle> contents = page.getPageContents();
	auto ph = PointerHolder<QPDFObjectHandle::StreamDataProvider>(new CombineFromContents_Provider(contents));
	xo_stream.replaceStreamData(ph, QPDFObjectHandle::newNull(), QPDFObjectHandle::newNull());
	return xo_stream;
}

/*** Formatting of coordinates ***/

string pdf_coord(double x, uint digits)
{
	char buf[16];
	snprintf(buf, sizeof(buf), "%.*f", digits, x);
	int n = strlen(buf);
	while (n > 0 && buf[n-1] == '0')
		buf[--n] = 0;
	if (n > 0 && buf[n-1] == '.')
		buf[--n] = 0;
	return buf;
}

/*** Rotation matrix construction ***/

pdf_matrix pdf_rotation_matrix(int deg, double width, double height)
{
      pdf_matrix m;
      switch (deg) {
	      case 0:
		      break;
	      case 90:
		      m.rotate_deg(-90);
		      m.shift(0, width);
		      break;
	      case 180:
		      m.rotate_deg(180);
		      m.shift(width, height);
		      break;
	      case 270:
		      m.rotate_deg(90);
		      m.shift(height, 0);
		      break;
	      default:
		      abort();
      }
      return m;
}
